pub mod functions;
pub mod layers;

use functions::loss_functions::{self, parse_type};
use layers::*;
use ndarray::{Array1, ArrayView1};

pub struct Network {
    layers: Vec<Box<dyn Layer>>,
    loss: fn(ArrayView1<f64>, ArrayView1<f64>) -> f64,
    loss_prime: fn(ArrayView1<f64>, ArrayView1<f64>) -> Array1<f64>
}

impl Network {
    pub fn new(loss_fn: loss_functions::Type) -> Self {
        let (loss, loss_prime) = parse_type(loss_fn);
        Network {
            layers: vec![],
            loss,
            loss_prime
        }
    }

    pub fn add_layer(&mut self, layer: Box<dyn Layer>) {
        self.layers.push(layer);
    }

    pub fn predict(&mut self, inputs: Vec<Array1<f64>>) -> Vec<Array1<f64>> {
        assert!(inputs.len() > 0);
        let mut result = vec![];

        for input in inputs.iter() {
            let mut output = Array1::default(inputs[0].raw_dim());
            output.assign(&input);
            for layer in &mut self.layers {
                output = layer.forward_pass(output.view());
            }
            result.push(output.to_owned());
        }

        result
    }

    pub fn fit(&mut self, x_train: Vec<Array1<f64>>, y_train: Vec<Array1<f64>>, epochs: usize, learning_rate: f64, trivial_optimize: bool) {
        assert!(x_train.len() > 0);
        assert!(x_train.len() == y_train.len());
        let num_samples = x_train.len();

        for i in 0..epochs {
            let mut err = 0.0;
            for j in 0..num_samples {
                // forward propagation
                let mut output = Array1::default(x_train[0].raw_dim());
                output.assign(&x_train[j]);
                for layer in self.layers.iter_mut() {
                    output = layer.forward_pass(output.view());
                }

                // compute loss
                err += (self.loss)(y_train[j].view(), output.view());

                // backward propagation
                let mut error = (self.loss_prime)(y_train[j].view(), output.view());
                for layer in self.layers.iter_mut().rev() {
                    if trivial_optimize {
                        error = layer.backward_pass(error.view(), learning_rate / (i+1) as f64);
                    } else {
                        error = layer.backward_pass(error.view(), learning_rate);
                    }
                }
            }
            // calculate average error on all samples
            err /= num_samples as f64;
            println!("epoch {}/{}  error={}", i+1, epochs, err);
        }
    }
}
